/*
 * ypauth: YP authentication DLL for Windows 2000
 * Copyright (C) Matthew Chapman 2001-2002
 */

#include "ypauth.h"

/* LSA function table */
extern PLSA_DISPATCH_TABLE			LSA;


/* Duplicate a SID */
PSID
  DuplicateSid(
    PSID Sid
  )
{
	DWORD NewSize = GetLengthSid(Sid);
	PSID NewSid;

	NewSid = LSA->AllocateLsaHeap(NewSize);
	CopySid(NewSize, NewSid, Sid);
	return NewSid;
}


/* Create a new SID consisting of DomainSid with Rid appended */
PSID
  AppendSubAuthority(
    PSID DomainSid,
    DWORD Rid
  )
{
	UCHAR SubAuths = *GetSidSubAuthorityCount(DomainSid);
	DWORD NewSize = GetSidLengthRequired((UCHAR)(SubAuths+1));
	PSID NewSid;

	NewSid = LSA->AllocateLsaHeap(NewSize);
	CopySid(NewSize, NewSid, DomainSid);
	(*GetSidSubAuthority(NewSid, SubAuths)) = Rid;
	(*GetSidSubAuthorityCount(NewSid))++;
	return NewSid;
}


/* Duplicate a UNICODE_STRING */
PUNICODE_STRING
  DuplicateUnicodeString(
    PUNICODE_STRING UnicodeString
  )
{
	PUNICODE_STRING NewString;

	NewString = LSA->AllocateLsaHeap(sizeof(UNICODE_STRING));
	NewString->Length = UnicodeString->Length;
	NewString->MaximumLength = UnicodeString->MaximumLength;
	NewString->Buffer = LSA->AllocateLsaHeap(NewString->MaximumLength);
	RtlCopyMemory(NewString->Buffer, UnicodeString->Buffer, NewString->Length);

	return NewString;
}


/* Copy SourceString into DestString allocating storage */
void
  CopyUnicodeString(
    PUNICODE_STRING DestString,
    PUNICODE_STRING SourceString
  )
{
	DestString->Length = SourceString->Length;
	DestString->MaximumLength = SourceString->MaximumLength;
	DestString->Buffer = LSA->AllocateLsaHeap(DestString->MaximumLength);
	RtlCopyMemory(DestString->Buffer, SourceString->Buffer, DestString->Length);
}


/* Extract password and UID from YP entry */
static char *
  ParsePasswdEntry(
    char *ValueData,
    int ValueLength,
    PUSHORT Uid
  )
{
	while ((ValueLength-- > 0) && (*(ValueData++) != ':'));

	if ((ValueLength <= CRYPT_LEN) || (ValueData[CRYPT_LEN] != ':'))
		return NULL;

	*Uid = (USHORT)atoi(&ValueData[CRYPT_LEN+1]);
	return ValueData;
}


/* Verify logon against YP and return UID */
NTSTATUS
  YpLogon(
    char *server,
    PSTRING UserName,
    PSTRING Password,
    PUSHORT Uid
  )
{
	CLIENT *clnt;
	struct ypreq_key req;
	struct ypresp_val resp;
	struct timeval timeout;
	char *cryptpw;

	clnt = clnt_create(server, YPPROG, YPVERS, "udp");
	if (clnt == NULL)
		return STATUS_NO_LOGON_SERVERS;

	req.domain = YP_DOMAIN;
	req.map = YP_MAP;
	req.key.keydat_len = UserName->Length;
	req.key.keydat_val = UserName->Buffer;
	timeout.tv_sec = 2;
	timeout.tv_usec = 0;
	RtlZeroMemory(&resp, sizeof(resp));

	if (clnt_call(clnt, YPPROC_MATCH, xdr_ypreq_key, &req, xdr_ypresp_val, &resp, &timeout) != RPC_SUCCESS)
	{
		clnt_destroy(clnt);
		return STATUS_NO_LOGON_SERVERS;
	}

	if (resp.stat == YP_TRUE)
	{
		cryptpw = ParsePasswdEntry(resp.val.valdat_val, resp.val.valdat_len, Uid);
		if ((cryptpw != NULL)
		    && (RtlCompareMemory(crypt(Password->Buffer, cryptpw), cryptpw, CRYPT_LEN) == CRYPT_LEN))
		{
			clnt_destroy(clnt);
			return STATUS_SUCCESS;
		}
	}

	clnt_destroy(clnt);
	return STATUS_LOGON_FAILURE;
}


/* Obtain information about the computer's account domain */
NTSTATUS
  InitializeDomain(
    POLICY_ACCOUNT_DOMAIN_INFO **Info
  )
{
	LSA_HANDLE LsaHandle;
	LSA_OBJECT_ATTRIBUTES ConnAttribs;
	NTSTATUS Status;

	RtlZeroMemory(&ConnAttribs, sizeof(ConnAttribs));
	Status = LsaOpenPolicy(NULL, &ConnAttribs, POLICY_VIEW_LOCAL_INFORMATION, &LsaHandle);
	if (Status != STATUS_SUCCESS)
		return Status;

	Status = LsaQueryInformationPolicy(LsaHandle, PolicyAccountDomainInformation, Info);
	if (Status != STATUS_SUCCESS)
		return Status;

	LsaClose(LsaHandle);
	return STATUS_SUCCESS;
}


/* Marshall a MSV1_0_INTERACTIVE_PROFILE into a client buffer */
void
  CreateProfile(
    PLSA_CLIENT_REQUEST ClientRequest,
    PVOID* ProfileBuffer,
    PULONG ProfileBufferSize
  )
{
	MSV1_0_INTERACTIVE_PROFILE *Profile;
	PCHAR BufferAddress, ClientAddress;

	*ProfileBufferSize = sizeof(*Profile) + LOGON_SERVER_LEN + LOGON_SCRIPT_LEN;
	Profile = LSA->AllocateLsaHeap(*ProfileBufferSize);
	LSA->AllocateClientBuffer(ClientRequest, *ProfileBufferSize, ProfileBuffer);
	RtlZeroMemory(Profile, *ProfileBufferSize);

	Profile->MessageType = MsV1_0InteractiveProfile;
	Profile->LogoffTime.LowPart
		= Profile->KickOffTime.LowPart
		= Profile->PasswordCanChange.LowPart
		= Profile->PasswordMustChange.LowPart = 0xffffffff;
	Profile->LogoffTime.HighPart
		= Profile->KickOffTime.HighPart
		= Profile->PasswordCanChange.HighPart
		= Profile->PasswordMustChange.HighPart = 0x7fffffff;

	BufferAddress = (PCHAR)Profile + sizeof(*Profile);
	ClientAddress = (PCHAR)*ProfileBuffer + sizeof(*Profile);

	Profile->LogonServer.Length = LOGON_SERVER_LEN - 2;
	Profile->LogonServer.MaximumLength = LOGON_SERVER_LEN;
	Profile->LogonServer.Buffer = (LPWSTR)ClientAddress;
	RtlCopyMemory(BufferAddress, LOGON_SERVER, LOGON_SERVER_LEN);

	BufferAddress += LOGON_SERVER_LEN;
	ClientAddress += LOGON_SERVER_LEN;

	Profile->LogonScript.Length = LOGON_SCRIPT_LEN - 2;
	Profile->LogonScript.MaximumLength = LOGON_SCRIPT_LEN;
	Profile->LogonScript.Buffer = (LPWSTR)ClientAddress;
	RtlCopyMemory(BufferAddress, LOGON_SCRIPT, LOGON_SCRIPT_LEN);

	LSA->CopyToClientBuffer(ClientRequest, *ProfileBufferSize, *ProfileBuffer, Profile);
	LSA->FreeLsaHeap(Profile);
}
